
// roamtest.l                              simple ROAM algorithm demo
// by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
//
// 03-29-02: wrote simple split-only
// 03-30-02: diamond cache; dynamic fractal terrain
// 04-01-02: full cache use, frustum cull,
//           better subdivision (-1 9 9 -1 tensor-product)
// 04-02-02: split roam library from this test application
// 07-15-02: updated to decorate the new spin menu system


lib_include "Spin/spin.l";
lib_include "../roam.l";


roam rm;
spin sp;
int done,nodraw;
double dtricntmax;


#define FRAME_NORMAL 0
#define FRAME_SAVE   1
#define FRAME_LOAD   2

int framecnt,framemode,framecntmax;
double framestore[0x100000];


void spevent(spin sp,int eventnum);
void do_draw();
void settricntmax();


lib_main {
    int x,y,sx,sy;

    framecnt=0;
    framemode=FRAME_NORMAL;
    dtricntmax=300000.0;
    nodraw=0;
    win_screensize(&sx,&sy);
    x=720; y=540;
    //x=y=128;
    printf("initial window size: x=%d y=%d\n",x,y);
    sp=spin_create("roamtest",sx/2-x/2,sy/2-y/2,x,y,spevent);
    {
        pack pk,pk1;
        spinmenu spm;

        pk=sp->pk_spm_default; spm=(spinmenu)pk->data;
        pk1=pk->pkc1;
        spinmenu_createkey(pk,'g',"g     -- grid view on");
        spinmenu_createkey(pk,'G',"G     -- grid view off");
        spinmenu_createkey(pk,'>',">     -- increase tri count");
        spinmenu_createkey(pk,'<',"<     -- decrease tri count");
        spinmenu_createkey(pk,'f',"f     -- freeze mesh");
        spinmenu_createkey(pk,'F',"F     -- unfreeze mesh");
        spinmenu_createkey(pk,'n',"n     -- skip mesh draw");
        spinmenu_createkey(pk,'N',"N     -- resume mesh draw");
        pack_add(pk,pk1); // put esc back at end
    }
    //spin_setnoclearswap(sp,1);
    spin_setper(sp,2.0,0.0,0.00001,4.0);
    rm=roam_create();
    settricntmax();
    for (done=0;!done;) sel_wait(300000);
};


void spevent(spin sp,int eventnum)
{
    win w;

    #define QQQ sqrt(sqrt(2.0))

    w=sp->w;
    switch (eventnum) {
    case WIN_DRAW:
        do_draw();
        break;
    case WIN_QUIT:
        done=1;
        break;
    case WIN_KEYDOWN:
        switch (w->keycode) {
        case 'g':
            roam_setdraw_grid(rm,1);
            break;
        case 'G':
            roam_setdraw_grid(rm,0);
            break;
        case '>':
            dtricntmax*=QQQ;
            settricntmax();
            break;
        case '<':
            dtricntmax/=QQQ;
            settricntmax();
            break;
        case 'f':
            roam_setfreeze(rm,1);
            break;
        case 'F':
            roam_setfreeze(rm,0);
            break;
        case 'n':
            nodraw=1;
            break;
        case 'N':
            nodraw=0;
            break;
        default:
            break;
        }
        break;
    default:
        break;
    }
}


void do_draw()
{
    double d;
    char msg[256];

    if (framemode!=FRAME_NORMAL) {
        static int firsttime=1;
        int k;
        double *dp;
        char *framename="framestore";

        dp=framestore+(framecnt*16);
        switch (framemode) {
        case FRAME_SAVE:
            for (k=0;k<16;k++) *dp++ =sp->vm->m[k>>2][k&3];
            fio_write(framename,(char *)framestore,framecnt*16*sizeof(double));
            break;
        case FRAME_LOAD:
            if (firsttime) {
                k=fio_size(framename);
                framecntmax=k/(16*sizeof(double));
                fio_setbuf((char *)framestore);
                fio_read(framename);
            }
            for (k=0;k<16;k++) sp->vm->m[k>>2][k&3]= *dp++;
            break;
        default:
            break;
        }
        framecnt++;
        if (framemode==FRAME_LOAD && framecnt>=framecntmax)
            framemode=FRAME_NORMAL;
        firsttime=0;
    }

    if (nodraw) {
        //spin_swap(sp); spin_clear(sp);
        return;
    }
    glColor3ub(255,255,255);
    roam_setfrustum(rm,sp->vm);
    d=rm->d_close;
    spin_setper(sp,2.0,0.0,0.7*d,4.0);
    //spin_swap(sp); spin_clear(sp);
    roam_draw(rm);
    sprintf(msg,"Mtri/s=%.2f %s",
        sp->fps*(double)(ROAM_PATCH_TRIS*rm->patchcnt)/1e6,rm->msg);
    spin_setmsg(sp,msg);
}

void settricntmax()
{
    int k;

    k=(int)floor(dtricntmax+0.5);
    if (k&(ROAM_PATCH_TRIS-1)) {
        k+=ROAM_PATCH_TRIS-(k&(ROAM_PATCH_TRIS-1));
    }
    k+=ROAM_PATCH_TRIS;
    roam_settricntmax(rm,k);
}


