
// packtest.l                                     test of packaging support
// by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
//
// 04-20-94: wrote
// 05-11-97: trying out the new features
// 05-25-97: extensive attachment/callback test using subtree sums with
//           lazy evaluation


lib_include "../pack.l";

void pki_event(pack pk,int eventnum);
void pki_randomize(pack pk,int level);
void pki_update(pack pk);
void pki_check(pack pk);

lib_main {

    // test of basic functionality (from 4-20-94)
    {
        pack pk,pk1,pk2,pkl;
        str buf,bufl;
 
        printf("\n\n");
        printf("========== simple pack create/save/load example ==========\n");
        pk=pack_create("--- ROOT ---");
            pk1=pack_create("1"); pack_add(pk,pk1);
                pk2=pack_create("1a"); pack_add(pk1,pk2);
                pk2=pack_create("1b"); pack_add(pk1,pk2);
                pk2=pack_create("1c"); pack_add(pk1,pk2);
            pk1=pack_create("2"); pack_add(pk,pk1);
                pk2=pack_create("2a"); pack_add(pk1,pk2);
                pk2=pack_create("2b"); pack_add(pk1,pk2);
                pk2=pack_create("2c"); pack_add(pk1,pk2);
            pk1=pack_create("3"); pack_add(pk,pk1);
                pk2=pack_create("3a"); pack_add(pk1,pk2);
                pk2=pack_create("3b"); pack_add(pk1,pk2);
                pk2=pack_create("3c"); pack_add(pk1,pk2);
        buf=pack_save(pk);
        printf("\n\npk =\n%s",buf);
        pkl=pack_load(buf);
        if (!pkl) {
            fprintf(stderr,"packtest: pack_load failed\n");
            fprintf(stderr,"*** %s\n",pack_error);
            exit(1);
        }
        bufl=pack_save(pkl);
        printf("\n\npkl =\n%s",bufl);
        str_destroy(buf); str_destroy(bufl);
        pack_destroy(pk); pack_destroy(pkl);
    }

    // try reading example.pk and copying it
    {
        char *buf;
        pack pk,pk1;

        printf("\n\n");
        printf("========== read and copy example.pk ==========\n");
        if (!(buf=fio_read("example.pk")))
            { fprintf(stderr,"*** can't read example.pk\n"); exit(1); }
        pk=pack_load(buf);
        if (!pk) {
            fprintf(stderr,"packtest: pack_load failed\n");
            fprintf(stderr,"*** %s\n",pack_error);
            exit(1);
        }
        mem_put(buf,"main");
        pack_print(pk);
        pk1=pack_copy(pk);
        pack_print(pk1);
        pack_destroy(pk);
        pack_destroy(pk1);
    }

    // example data+callback attachments: subtree-sum int's
    {
        int i;
        pack pk,pk1;
        char *buf;

        // register event handler for LOAD processing
        // Note: this is needed because the pack system doesn't know
        // about events unless given by the application, and this
        // information is not carried in a saved pack.
        pack_load_event=pki_event;

        pk=pack_createdata("root",(char *)0,pki_event);
        pack_print(pk);
        for (i=0;i<1500;i++) {
            printf("."); fflush(stdout);
            pki_randomize(pk,0);
            pki_check(pk);
        }
        printf("\n");
        pack_print(pk);
        pk1=pack_copy(pk);
        pki_check(pk1);
        pack_print(pk1);
        pack_destroy(pk1);
        buf=pack_save(pk);
        pk1=pack_load(buf);
        if (!pk1) {
            fprintf(stderr,"packtest: pack_load failed\n");
            fprintf(stderr,"*** %s\n",pack_error);
            exit(1);
        }
        pki_check(pk1);
        pack_print(pk1);
        pack_destroy(pk1);
        pki_update(pk);
        pack_print(pk);
        pack_destroy(pk);
    }
};

#define PKI_DIRTY 0x80000000

void pki_event(pack pk,int eventnum)
{
    pack pkp;
    int *ip,*ips,n;
    char buf[256],*d;

    ip=(int *)pk->data;
    switch (eventnum) {
    case  PACK_EVENT_CREATE:
        if (!pk->data) {
            pk->data=mem_get(sizeof(int),"pki_event");
            ip=(int *)pk->data; *ip=0;
        }
        break;
    case  PACK_EVENT_DESTROY:
        if (pk->data) mem_put(pk->data,"pki_event");
        break;
    case  PACK_EVENT_COPY:
        ips=(int *)pack_copy_pk->data;
        *ip= *ips;
        break;
    case  PACK_EVENT_DELETE:
    case  PACK_EVENT_INSERT:
        for (pkp=pk->pkp;pkp && pkp->event==pki_event &&
            *(ips=(int *)pkp->data)!=PKI_DIRTY;pkp=pkp->pkp) {
            *ips=PKI_DIRTY;
        }
        break;
    case  PACK_EVENT_SAVE:
        // automatic update on SAVE (un-comment if desired)
        // if (!pk->pkp || pk->pkp->event!=pki_event) pki_update(pk);
        if (*ip!=PKI_DIRTY) sprintf(buf,"%d",*ip);
        else sprintf(buf,"?");
        pack_save_pk=pack_create(buf);
        break;
    case  PACK_EVENT_LOAD:
        pk->event=pki_event;
        pki_event(pk,PACK_EVENT_CREATE); ip=(int *)pk->data;
        n=0;
        if (pack_load_pk) {
            d=pack_load_pk->description;
            if (d[0]=='?') n=PKI_DIRTY;
            else sscanf(pack_load_pk->description,"%d",&n);
        }else{
            fprintf(stderr,"*** pki_event LOAD: missing attachment pack\n");
            exit(1);
        }
        if (!pk->pkc0 && n==PKI_DIRTY) n=0;
        *ip=n;
        break;
    default:
        printf("pki_event: Eeek! Unknown pack event type!!!!  num=%d\n",
            eventnum);
        break;
    }
}

void pki_randomize(pack pk,int level)
{
    int r,*ip;
    pack pkc,pkc1,pkp;

    for (pkc=pk->pkc0;pkc;pkc=pkc1)
        { pkc1=pkc->pkl1; pki_randomize(pkc,level+1); }

    r=RNDi&0xf;
    switch (r) {
    case 0: // create
    case 1: // create
        if (level<10) {
            pkc=pack_createdata("pki",(char *)0,pki_event);
            r=RNDi&7;
            ip=(int *)pkc->data;
            *ip=r;
            pack_add(pk,pkc);
        }
        break;
    case 2: // destroy
        if ((pkp=pk->pkp)) {
            pack_destroy(pk);
            if (!pkp->pkc0) {
                ip=(int *)pkp->data;
                *ip=RNDi&7;
            }
        }
        break;
    case 3: // move (delete+insert)
        if (pk->pkc0) pack_add(pk,pk->pkc0);
        break;
    default:
        break;
    }
}

void pki_update(pack pk)
{
    int *ip,*ipc,n;
    pack pkc;

    ip=(int *)pk->data;
    if (*ip!=PKI_DIRTY) return;
    n=0;
    for (pkc=pk->pkc0;pkc;pkc=pkc->pkl1) {
        pki_update(pkc);
        ipc=(int *)pkc->data;
        n+= *ipc;
    }
    *ip=n;
}

void pki_check(pack pk)
{
    int *ip,*ipc,n,dirty;
    pack pkc;

    ip=(int *)pk->data; n=0; dirty=0;
    if (pk->pkc0) {
        for (pkc=pk->pkc0;pkc;pkc=pkc->pkl1) {
            pki_check(pkc);
            ipc=(int *)pkc->data;
            if (*ipc==PKI_DIRTY) dirty++;
            else n+= *ipc;
        }
        if (dirty && *ip!=PKI_DIRTY) {
            fprintf(stderr,"*** pki_check: dirty not propigated\n");
        }
        if (*ip!=PKI_DIRTY && *ip!=n) {
            fprintf(stderr,"*** pki_check: bad sum %d should be %d\n",*ip,n);
            pack_print(pk);
        }
    }else{
        if (*ip==PKI_DIRTY) {
            fprintf(stderr,"*** pki_check: dirty leaf\n");
        }
    }
}


